# import.bashの解説

2019.7.14 Masahiro Nakajima

Software Design 2019年9月号 Unixコマンドライン探検隊の、Redis解説記事で用いた、辞書データをRedisにインポートするスクリプトimport.bashについて解説します。以下、すでに誌面で記述した内容の詳細は記述しませんので、記事を参照しながら読み進めてください。

import.bashでは、redis-cliとの連携方法、テキストデータの加工のTIPSの習得、そして何よりも「実際に現場でデータインポートする場合の試行錯誤しながら作業するためのテンプレート」として使うことを想定しています。

# 基本的な使い方

import.bashのオプションは、以下のとおりです。

・f…… Redisのデータベースをクリアする
・p…… 対話的に確認しながら実行
・v…… 冗長メッセージ表示(デバッグ用)
・i…… redis-cliをスクリプト内で呼び出す
・P…… redis-cli --pipe で大量データをインサートする
・l…… インポートする件数の上限
・L…… 進捗ログメッセージの出力先
・e, j…… 各辞書ファイルの場所

実行にあたっては、辞書データを用意(UTF-8の標準的なUnixテキスト形式)、配置します。辞書の配置場所が、デフォルトのパス・ファイル名と異なる場合は、上記オプションで場所を示してください。

以下、既存のデータを削除し、新規にデータをロードする3つの方法を示します。

1. パイプで接続

```console
$ ./import.bash -f | redis-cli
```

もっとも基本的な使い方。とにかくデータをロードして、辞書の動作を確認したいなら、この方法でよいでしょう。import.bash, redis-cliそれぞれの処理結果出力を画面に表示させると、処理が遅くなるので、以下のように実行するのが基本です。

```console
# ログが不要
$ ./import.bash -f -L /dev/null | redis-cli > /dev/null
```

```console
# import.bash, redis-cli それぞれ、別のファイルに結果を出力しておく
$ ./import.bash -f -L logfile_import | redis-cli > logfile_redis
```

2. import.bash内部からredis-cliを呼び出す

```console
$ ./import.bash -fi
```

import.bash内から、1レコードごとにredis-cliコマンドを実行して、Redisコマンドを処理します。当然、処理負荷が高く決して懸命なやりかたではありません。デバッグ目的などで、性能よりも、1行ごとの処理結果により、続く処理をどうするか、などの対処を丁寧に実施したい場合に有用です。

3. 大量のデータを高速でインサートする場合

```console
$ ./import.bash -fP -L /dev/null > ejdict_redis_protocol.txt
$ redis-cli --pipe < ejdict_redis_protocol.txt 
```

Redisに大量データを挿入する場合に、効果的な方法がこのやりかたです。1度きりのロードなら、単純に以下のように処理しても、同じ結果が得られ、1.と変わりはありません。プロトコルデータを作成する優位性については、後述の性能比較を参照してください。

```console
$ ./import.bash -fP -L /dev/null | redis-cli --pipe
```

## 標準出力・標準エラー出力・ファイルへのリダイレクト

redis-cliは標準出力からの入力を処理します。パイプで接続する場合は、Redisコマンド以外は、標準エラー出力かファイルへ出力してやらなければなりません。これは、-p オプションを使って、対話的に操作の継続を促す場合にも重要です。オプションで、進捗ログ、冗長メッセージの出力先を変更できるようにしてあるのは、このためです。

また、パイプで接続した際は、redis-cliコマンドの標準出力や標準エラー出力の出力もあります。これらの情報も、適切に処理する必要がありいます。

## -l, -p オプションを上手に使って少しずつ

大きなデータをロードするには、時間がかかり、データ領域も消費します。また通常、はじめて処理するデータ形式にたいしては、データの加工やローディングの仕方が妥当かを試行錯誤しながら、スクリプトを修正してロードを繰り返すことになります。このような場合、一度に元のデータをすべてインポートしてしまうのではなく、最初は数〜数十レコードのみをロードして、動作を確認するのが賢いやり方です。import.bashでは、そのために -l オプションを用意しています。

インポートのやり直しをする際は、-f オプションでRedis内のデータをクリアできます。また、-p オプションで、要所で処理を続行するか対話的に作業を進めることができます。

注意: 今回の辞書だけで使っているRedisサーバでない限り、-fオプションでのクリアは危険です。他のデータベースも削除してしまうことがありますので、注意してください。

# 大量データのインサート機能の性能検証

Redisに大量にデータを流し込む場合、高速に安全に処理するためのプロトコル( [Redis Protocol specification] )があります。Rubyなどの外部プログラムから、このプロトコルに従って、一気にデータをロードすることができます( [大量データのインサート] )。

redis-cliは、このプロトコルを用いてサーバと通信する --pipe オプションが備えられています。使い方は、先例の**基本的な使い方**で用いた、パイプで接続する方式とほぼ同様ですが、redis-cliに送るRedisコマンドの行区切りをUnixスタイルの<LF>ではなく、<CR><LF>(DOSスタイル)にして送る必要があります。import.bashでは、-P オプションを指定することで、この形式でRedisコマンドを標準出力に出力します。

[大量データのインサート]: https://redis-documentasion-japanese.readthedocs.io/ja/latest/topics/mass-insert.html
[Redis Protocol specification]: https://redis.io/topics/protocol

## redis-cli --pipe による性能効果を確認する

測定の前に、インポートしたデータのレコード数を確認するコマンドを見ておきます。誌面で紹介したデータを使って正しくロードできれば、以下のような結果(79,384レコード)になるはずです。以降、いくつか異なる方法で、データをロードしますが、実施語少なくともレコード件数が正しいかの確認は、しておきましょう。

```shell-session
# ロードしたデータのレコード数を確認する
$ redis-cli --scan --pattern "*" | wc -l
79384
```

>
以下、計測した実行時間は、筆者の使用しているUbuntu 18.04LTS搭載の同一マシン上での結果です。いずれも、1回ずつの計測で、統計的な処理は行っていませんので、誤差・ばらつきによる変動があります。

import.bashからパイプで直接redis-cliを接続した場合、通常のリダイレクトによる処理(9m 4.757s) と --pipeでの処理で(9m 10.199sec)に大きな差が見られません。

```console
# 通常のリダイレクト
$ time ./import.bash -f -L /dev/null | redis-cli > /dev/null
```

```console
# import.bashからredis-cli --pipe にリダイレクトする
$ time ./import.bash -fP -L /dev/null | redis-cli --pipe
```

そこで、import.bashによるデータ処理時間のみ計測します。--pipe無しのRedisコマンドの羅列を生成した場合(9m 38.687s)、--pipeによるRedis プロトコルデータの生成(9m 50.951s) と、ほぼBashスクリプトの処理時間が、インポートの時間の大半を占めていることがわかります。

```console
# Redisコマンドの羅列ファイルを生成する
$ time ./import.bash -f -L /dev/null > ejdict_standard.txt
```

```console
# パイプライン処理用にプロプロトコルデータを出力する
$ time ./import.bash -fP -L /dev/null > ejdict_redis_protocol.txt
```

続いて、生成したファイルをredis-cliに送り、ロード時間を計測します。--pipeなしで 3.647s、--pipeで 0.830sと、--pipeオプションが高速であることが確認できました。

データを繰り返しインポートすることがあったり、本ケースよりもはるかに大きなデータをロードする場合には、前処理したデータを用意して、--pipeオプションを用いて処理するのが効果的です。

```console
# 単純なコマンドの羅列をredis-cliに投入する
$ time redis-cli < ejdict_standard.txt > /dev/null 2>&1
```

```console
# プロトコルデータをインポートする
$ time redis-cli --pipe < ejdict_redis_protocol.txt
```

ちなみに、-iオプションの内部呼び出しでは、11m18.109s かかっており、プロセスの起動によるオーバヘッドが確認できます。

```shell-session
# import.bash内部からredis-cliを起動している場合
$ time ./import.bash -fi > /dev/null 2>&1
```
